define(['exports', 'query-string', 'normalizr', 'reselect', 'redux', 'react', 'hoist-non-react-statics', 'validate.js'], function (exports, queryString, normalizr, reselect, redux, React, hoistStatics, validatejs) { 'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  var queryString__default = /*#__PURE__*/_interopDefaultLegacy(queryString);
  var React__default = /*#__PURE__*/_interopDefaultLegacy(React);
  var hoistStatics__default = /*#__PURE__*/_interopDefaultLegacy(hoistStatics);
  var validatejs__default = /*#__PURE__*/_interopDefaultLegacy(validatejs);

  const NAMESPACE = 'snowbox';
  const actions = {
    UPSERT: `${NAMESPACE}/UPSERT`,
    REMOVE: `${NAMESPACE}/REMOVE`,
    FIND: `${NAMESPACE}/FIND`,
    FETCH: `${NAMESPACE}/FETCH`,
    CLEAR: `${NAMESPACE}/CLEAR`,
    NO_FETCH: `${NAMESPACE}/NO_FETCH`
  };
  const statuses = {
    PENDING: 'pending',
    SUCCEEDED: 'succeeded',
    FAILED: 'failed'
  };
  const contentTypes = {
    JSON: `${NAMESPACE}/JSON`,
    FORM_DATA: `${NAMESPACE}/FORM_DATA`
  };

  class Api {
    setBaseUrl(baseUrl) {
      if (typeof baseUrl != 'string') {
        throw new Error('[Snowbox] Api base url must be string');
      }

      this.baseUrl = baseUrl;
    }

    setTokenHeaderName(tokenHeader) {
      if (typeof tokenHeader != 'string') {
        throw new Error('[Snowbox] Token Header Name must be string');
      }

      this.tokenHeader = tokenHeader;
    }

    setHeaderRequest(name, value) {
      if (!this.requestHeaders) {
        this.requestHeaders = {};
      }

      if (typeof name != 'string') {
        throw new Error('[Snowbox] Request header name must be a string');
      }

      if (typeof value != 'string') {
        throw new Error('[Snowbox] Request header value must be a string');
      }

      this.requestHeaders[name] = value;
    }

    setTokenGetter(tokenGetter) {
      if (typeof tokenGetter != 'function') {
        throw new Error('[Snowbox] Token Getter must be function');
      }

      this.tokenGetter = tokenGetter;
    }

    async setAuthToken(xhr) {
      if (!this.tokenGetter || !this.tokenHeader) {
        return;
      }

      const token = await this.tokenGetter();

      if (!token) {
        return;
      }

      xhr.setRequestHeader(this.tokenHeader, token);
    }

    get(path, params) {
      return this.call('GET', path, params);
    }

    post(path, data = {}, params, contentType) {
      return this.call('POST', path, params, data, contentType);
    }

    put(path, data = {}, params, contentType) {
      return this.call('PUT', path, params, data, contentType);
    }

    patch(path, data = {}, params, contentType) {
      return this.call('PATCH', path, params, data, contentType);
    }

    remove(path) {
      return this.call('DELETE', path);
    }

    async call(method, path, params, data, contentType = contentTypes.JSON) {
      if (!this.baseUrl) {
        throw new Error('[Snowbox] Base API url must be defined');
      }

      return new Promise(async (resolve, reject) => {
        const xhr = new XMLHttpRequest();
        const sep = path.indexOf('?') > -1 ? '&' : '?';
        const qs = typeof params == 'object' ? sep + queryString__default['default'].stringify(params) : '';
        xhr.open(method, `${this.baseUrl}${path}${qs}`, true);

        xhr.onload = () => {
          if (xhr.readyState != 4) {
            return false;
          }

          let response;

          try {
            response = JSON.parse(xhr.responseText);
          } catch (err) {
            response = xhr.responseText;
          }

          if (xhr.status >= 200 && xhr.status < 300) {
            resolve(response.data ? response.data : response);
          } else {
            reject({
              status: xhr.status,
              message: response.message || response
            });
          }
        };

        let body;

        switch (contentType) {
          case contentTypes.JSON:
            xhr.setRequestHeader('Content-type', 'application/json; charset=utf-8');

            if (data) {
              body = JSON.stringify(data);
            }

            break;

          case contentTypes.FORM_DATA:
            xhr.setRequestHeader('Content-type', 'multipart/form-data');

            if (typeof data == 'object') {
              body = new FormData();
              Object.keys(data).forEach(field => body.append(field, data[field]));
            }

            break;

          default:
            return reject(new Error(`[Snowbox] Invalid content type "${contentType}"`));
        }

        if (this.requestHeaders && Object.keys(this.requestHeaders).length !== 0) {
          Object.keys(this.requestHeaders).forEach(requestHeader => {
            xhr.setRequestHeader(requestHeader, this.requestHeaders[requestHeader]);
          });
        }

        await this.setAuthToken(xhr);
        xhr.send(data ? body : undefined);
      });
    }

  }

  var api = new Api();

  class Provider {
    constructor(particleDefinition) {
      this.particleDefinition = particleDefinition;
    }

    fetch(filter) {
      const path = this.getFetchPath(filter);
      const params = this.getFetchParams(filter);
      return api.get(path, params);
    }

    upsert(data, params) {
      const path = this.getUpsertPath(data);
      const method = this.getUpsertMethod(data);
      return api[method](path, data, params, this.getUpsertContentType());
    }

    remove(data) {
      const particle = this.getParticle(data);
      const id = typeof data == 'number' ? data : data.id;
      return api.remove(`/${particle}/${id}`);
    }

    getParticle(data) {
      return typeof this.particleDefinition == 'function' ? this.particleDefinition(data) : this.particleDefinition;
    }

    getFetchPath(filter) {
      const particle = this.getParticle(filter);
      return !filter || typeof filter == 'object' ? `/${particle}` : `/${particle}/${filter}`;
    }

    getFetchParams(filter) {
      return typeof filter == 'object' ? filter : null;
    }

    getUpsertPath(data) {
      const particle = this.getParticle(data);
      return data.id ? `/${particle}/${data.id}` : `/${particle}`;
    }

    getUpsertMethod(data) {
      return data.id ? 'put' : 'post';
    }

    getUpsertContentType() {
      return contentTypes.JSON;
    }

  }

  const entity = (key, provider, relations = {}, options = {}) => {
    const newEntity = new normalizr.schema.Entity(key, relations, options);

    if (provider) {
      if (typeof provider != 'object' && provider.constructor.name != 'Provider') {
        throw new Error('[Snowbox] Invalid provider');
      }

      newEntity.provider = provider;
    }

    if (typeof options.staleTimeout == 'number') {
      newEntity.staleTimeout = options.staleTimeout;
    }

    if (options.singleton === true) {
      newEntity.singleton = true;
    }

    return newEntity;
  };

  const success = str => `${str}_SUCCESS`;
  const failure = str => `${str}_FAILURE`;
  const createAction = type => (data = {}) => ({
    type,
    ...data
  });
  const createSuccess = type => entity => (data, entities, result, date) => ({
    type: success(type),
    entity,
    data,
    entities,
    result,
    date
  });
  const createFailure = type => entity => (data, error, statusCode) => ({
    type: failure(type),
    entity,
    data,
    error,
    statusCode
  });
  const request = type => entity => (data, options = {}) => ({
    type,
    entity,
    data,
    options,
    success: createSuccess(type)(entity),
    failure: createFailure(type)(entity)
  });
  const upsert = request(actions.UPSERT);
  const remove = request(actions.REMOVE);
  const find = request(actions.FIND);
  const fetch = request(actions.FETCH);
  const clearAll = createAction(actions.CLEAR);
  const noFetch = createAction(actions.NO_FETCH);

  const baseKey = '#';
  const buildKey = (filters = {}) => {
    if (!filters || typeof filters != 'object' || Array.isArray(filters)) {
      return baseKey;
    }

    return Object.keys(filters).sort().reduce((key, filter) => {
      key += `${filter}[${filters[filter]}]`;
      return key;
    }, baseKey);
  };

  const selectSnowbox = state => state.snowbox;

  const selectMeta = reselect.createSelector(selectSnowbox, state => state.meta);
  const selectEntities = reselect.createSelector(selectSnowbox, state => state.entities);
  const selectSingletons = reselect.createSelector(selectSnowbox, state => state.singletons);

  const applyHydration = (entity, levels = 0) => entities => data => {
    if (!data || !levels) {
      return data;
    }

    if (Array.isArray(data)) {
      return data.map(item => applyHydration(entity, levels)(entities)(item));
    }

    const newData = { ...data
    };

    for (const key in entity.schema) {
      const childEntity = Array.isArray(entity.schema[key]) ? entity.schema[key][0] : entity.schema[key];
      newData[key] = applyHydration(childEntity, levels - 1)(entities)(Array.isArray(data[key]) ? data[key].map(id => entities[childEntity.key][id]) : entities[childEntity.key][data[key]]);
    }

    return newData;
  };

  const selectOne = (entity, hydrationLevels = 0) => id => {
    if (entity.singleton) {
      return reselect.createSelector(selectSingletons, singletons => singletons[entity.key]);
    }

    const hydrate = applyHydration(entity, hydrationLevels);
    return reselect.createSelector(selectEntities, entities => entities[entity.key] ? hydrate(entities)(entities[entity.key][id]) : null);
  };
  const emptyArray = [];
  const selectMany = (entity, hydrationLevels = 0) => filters => {
    const hydrate = applyHydration(entity, hydrationLevels);
    return reselect.createSelector(selectEntities, selectMeta, (allEntities, allMeta) => {
      if (!allEntities[entity.key]) {
        return emptyArray;
      }

      const key = buildKey(filters);
      const meta = allMeta[entity.key] ? allMeta[entity.key][key] : null;

      if (!meta || !meta.result) {
        return hydrate(allEntities)(Object.values(allEntities[entity.key]));
      }

      const selectedEntities = meta.result.map(id => allEntities[entity.key][id]);
      return hydrate(allEntities)(selectedEntities);
    });
  };

  const snowboxActions = [actions.FETCH, success(actions.FETCH), failure(actions.FETCH), actions.FIND, success(actions.FIND), failure(actions.FIND), actions.UPSERT, success(actions.UPSERT), failure(actions.UPSERT), actions.REMOVE, success(actions.REMOVE), failure(actions.REMOVE)];

  const entitiesReducer = (state = {}, action) => {
    var _action$entity;

    if (!snowboxActions.includes(action.type)) {
      return state;
    }

    if (action !== null && action !== void 0 && (_action$entity = action.entity) !== null && _action$entity !== void 0 && _action$entity.singleton) {
      return state;
    }

    if (action.type == success(actions.REMOVE)) {
      const newState = { ...state,
        [action.entity.key]: { ...state[action.entity.key]
        }
      };
      const id = typeof action.data == 'number' ? action.data : action.data[action.entity.idAttribute];
      delete newState[action.entity.key][id];
      return newState;
    }

    if (!action.entities) {
      return state;
    }

    const newState = { ...state
    };

    for (const entity in action.entities) {
      newState[entity] = { ...(newState[entity] || {})
      };

      for (const id in action.entities[entity]) {
        newState[entity][id] = { ...((newState[entity] || {})[id] || {}),
          ...action.entities[entity][id],
          __updatedAt: action.date
        };
      }
    }

    return newState;
  };

  const fetchEntityhMetaReducer = (state = {}, action) => {
    switch (action.type) {
      case actions.FETCH:
        return {
          progress: statuses.PENDING
        };

      case success(actions.FETCH):
        return {
          progress: statuses.SUCCEEDED,
          result: action.result,
          __updatedAt: action.date
        };

      case failure(actions.FETCH):
        return {
          progress: statuses.FAILED,
          error: action.error
        };
    }
  };

  const entityMetaReducer = (state = {}, action) => {
    const key = buildKey(action.data);
    return { ...state,
      [key]: fetchEntityhMetaReducer(state[key], action)
    };
  };

  const metaReducer = (state = {}, action) => {
    const allowedActions = [actions.FETCH, success(actions.FETCH), failure(actions.FETCH)];

    if (!allowedActions.includes(action.type)) {
      return state;
    }

    const entity = Array.isArray(action.entity) ? action.entity[0] : action.entity;

    if (entity.singleton) {
      return state;
    }

    return { ...state,
      [entity.key]: entityMetaReducer(state[entity.key], action)
    };
  };

  const singletonsReducer = (state = {}, action) => {
    var _action$entity2;

    if (!snowboxActions.includes(action.type)) {
      return state;
    }

    if (!(action !== null && action !== void 0 && (_action$entity2 = action.entity) !== null && _action$entity2 !== void 0 && _action$entity2.singleton)) {
      return state;
    }

    const newState = { ...state
    };

    if (action.type == success(actions.REMOVE)) {
      delete newState[action.entity.key];
      return newState;
    }

    newState[action.entity.key] = { ...action.result,
      __updatedAt: action.date
    };
    return newState;
  };

  const rootReducer = redux.combineReducers({
    entities: entitiesReducer,
    meta: metaReducer,
    singletons: singletonsReducer
  });
  const snowboxReducer = (state, action) => {
    if (action.type == actions.CLEAR) {
      return rootReducer(undefined, action);
    }

    return rootReducer(state, action);
  };

  const methods = {
    [actions.UPSERT]: 'upsert',
    [actions.REMOVE]: 'remove',
    [actions.FIND]: 'fetch',
    [actions.FETCH]: 'fetch'
  };
  const shouldFetchData = (state, entity, action) => {
    var _action$options, _stateData;

    if (!entity.staleTimeout) {
      return true;
    }

    if (action !== null && action !== void 0 && (_action$options = action.options) !== null && _action$options !== void 0 && _action$options.refresh) {
      return true;
    }

    let stateData;

    if (Array.isArray(action.entity)) {
      var _state$snowbox, _state$snowbox$meta, _state$snowbox$meta$e;

      const key = buildKey(action.data);
      stateData = state === null || state === void 0 ? void 0 : (_state$snowbox = state.snowbox) === null || _state$snowbox === void 0 ? void 0 : (_state$snowbox$meta = _state$snowbox.meta) === null || _state$snowbox$meta === void 0 ? void 0 : (_state$snowbox$meta$e = _state$snowbox$meta[entity.key]) === null || _state$snowbox$meta$e === void 0 ? void 0 : _state$snowbox$meta$e[key];
    } else if (entity.singleton) {
      var _state$snowbox2, _state$snowbox2$singl;

      stateData = state === null || state === void 0 ? void 0 : (_state$snowbox2 = state.snowbox) === null || _state$snowbox2 === void 0 ? void 0 : (_state$snowbox2$singl = _state$snowbox2.singletons) === null || _state$snowbox2$singl === void 0 ? void 0 : _state$snowbox2$singl[entity.key];
    } else {
      var _state$snowbox3, _state$snowbox3$entit, _state$snowbox3$entit2;

      const id = action.data[entity.idAttribute];
      stateData = state === null || state === void 0 ? void 0 : (_state$snowbox3 = state.snowbox) === null || _state$snowbox3 === void 0 ? void 0 : (_state$snowbox3$entit = _state$snowbox3.entities) === null || _state$snowbox3$entit === void 0 ? void 0 : (_state$snowbox3$entit2 = _state$snowbox3$entit[entity.key]) === null || _state$snowbox3$entit2 === void 0 ? void 0 : _state$snowbox3$entit2[id];
    }

    if (!((_stateData = stateData) !== null && _stateData !== void 0 && _stateData.__updatedAt)) {
      return true;
    }

    return Date.now() - stateData.__updatedAt > entity.staleTimeout;
  };
  const snowboxMiddleware = store => next => async action => {
    if (![actions.UPSERT, actions.REMOVE, actions.FIND, actions.FETCH].includes(action.type)) {
      return next(action);
    }

    const method = methods[action.type];
    const entity = Array.isArray(action.entity) ? action.entity[0] : action.entity;

    if (method == 'fetch' && !shouldFetchData(store.getState(), entity, action)) {
      return next(noFetch());
    }

    next(action);

    try {
      const response = await entity.provider[method](action.data);

      if (method == 'remove') {
        return next(action.success(action.data));
      }

      if (entity.singleton) {
        return next(action.success(action.data, undefined, response, Date.now()));
      }

      const {
        entities,
        result
      } = normalizr.normalize(response, action.entity);
      return next(action.success(action.data, entities, result, Date.now()));
    } catch (error) {
      return next(action.failure(action.data, error, error.status));
    }
  };

  const buildInitialState = fieldList => (hocInitialValues = {}, propsInitialValues) => {
    const state = {
      valid: true,
      touched: false,
      error: null,
      submitted: false,
      fields: {}
    };
    const values = propsInitialValues ? propsInitialValues : hocInitialValues;
    fieldList.forEach(field => {
      state.fields[field] = {
        valid: true,
        touched: false,
        error: null,
        value: typeof values[field] == 'undefined' ? null : values[field]
      };
    });
    return state;
  };
  const getData = fieldList => state => {
    return fieldList.reduce((data, field) => ({ ...data,
      [field]: state.fields[field].value
    }), {});
  };
  const validate = (fieldList, constraints) => state => {
    const errors = validatejs__default['default'](getData(fieldList)(state), constraints);
    let isFormValid = true;
    fieldList.forEach(field => {
      const error = errors && errors[field] ? errors[field][0] : null;
      isFormValid = isFormValid && !error;
      state = { ...state,
        valid: isFormValid,
        fields: { ...state.fields,
          [field]: { ...state.fields[field],
            error,
            valid: !error
          }
        }
      };
    });
    return state;
  };
  const handleChange = (fieldList, constraints) => (state, field, value) => {
    if (!fieldList.includes(field)) {
      throw new Error(`[Snowbox Form] Unknown field: ${field}`);
    }

    let newState = { ...state,
      fields: { ...state.fields,
        [field]: { ...state.fields[field],
          value
        }
      }
    };
    return validate(fieldList, constraints)(newState);
  };
  const handleBlur = fieldList => (state, field) => {
    if (!fieldList.includes(field)) {
      throw new Error(`[Snowbox Form] Unknown field: ${field}`);
    }

    return { ...state,
      touched: true,
      fields: { ...state.fields,
        [field]: { ...state.fields[field],
          touched: true
        }
      }
    };
  };
  const handleServerErrors = fieldList => (state, errors) => {
    const newState = fieldList.reduce((acc, field) => ({ ...acc,
      valid: acc.valid && !errors[field],
      submitted: false,
      fields: { ...acc.fields,
        [field]: { ...acc.fields[field],
          error: errors[field] || null,
          valid: !errors[field]
        }
      }
    }), state);

    if (errors._error) {
      newState.error = errors._error;
      newState.valid = false;
    }

    return newState;
  };
  const handleSubmit = fieldList => state => {
    const newState = fieldList.reduce((acc, field) => ({ ...acc,
      fields: { ...acc.fields,
        [field]: { ...acc.fields[field],
          touched: true
        }
      }
    }), state);

    if (newState.valid) {
      newState.submitted = true;
    }

    return newState;
  };

  const createFormService = (fieldList, constraints) => ({
    buildInitialState: buildInitialState(fieldList),
    getData: getData(fieldList),
    validate: validate(fieldList, constraints),
    handleChange: handleChange(fieldList, constraints),
    handleBlur: handleBlur(fieldList),
    handleServerErrors: handleServerErrors(fieldList),
    handleSubmit: handleSubmit(fieldList)
  });

  const withForm = ({
    fields: fieldConstraints,
    submitForm,
    initialValues = {}
  }) => WrappedComponent => {
    const fieldList = Object.keys(fieldConstraints);
    const formService = createFormService(fieldList, fieldConstraints);

    const SnowForm = (ownProps = {}) => {
      const [formState, setState] = React.useState(formService.buildInitialState(initialValues, ownProps.initialValues));
      const firstRender = React.useRef(true);
      React.useEffect(() => {
        if (firstRender.current) {
          firstRender.current = false;
          setState(state => formService.validate(state));
        }
      });

      const onChange = (field, value) => {
        setState(state => formService.handleChange(state, field, value));
      };

      const onBlur = field => {
        setState(state => formService.handleBlur(state, field));
      };

      const setServerErrors = errors => {
        setState(state => formService.handleServerErrors(state, errors));
      };

      const onSubmit = async () => {
        const newState = formService.handleSubmit(formState);
        setState(newState);

        if (!newState.valid) {
          return;
        }

        const resetState = await submitForm(ownProps, formService.getData(newState), setServerErrors);

        if (resetState !== false) {
          setState(formService.buildInitialState(initialValues, ownProps.initialValues));
          firstRender.current = true;
        }
      };

      const fieldsProp = Object.keys(formState.fields).reduce((fields, name) => ({ ...fields,
        [name]: { ...formState.fields[name],
          name,
          onBlur: () => onBlur(name),
          onChange: value => onChange(name, value)
        }
      }), {});
      return /*#__PURE__*/React__default['default'].createElement(WrappedComponent, { ...ownProps,
        fields: fieldsProp,
        form: {
          valid: formState.valid,
          touched: formState.touched,
          error: formState.error
        },
        onFieldChange: onChange,
        onFieldBlur: onBlur,
        onSubmit
      });
    };

    return hoistStatics__default['default'](SnowForm, WrappedComponent);
  };

  exports.Provider = Provider;
  exports.actions = actions;
  exports.api = api;
  exports.clearAll = clearAll;
  exports.contentTypes = contentTypes;
  exports.entity = entity;
  exports.fetch = fetch;
  exports.find = find;
  exports.remove = remove;
  exports.selectMany = selectMany;
  exports.selectOne = selectOne;
  exports.snowboxMiddleware = snowboxMiddleware;
  exports.snowboxReducer = snowboxReducer;
  exports.upsert = upsert;
  exports.withForm = withForm;

  Object.defineProperty(exports, '__esModule', { value: true });

});
